package border_finder;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.ArrayList;
import java.util.Random;

public class border_finder extends Frame {

	private ArrayList<Shape> circle_list = new ArrayList<Shape>();
	private ArrayList<Shape> border_list = new ArrayList<Shape>();
	private ArrayList<Shape> lowMin_list = new ArrayList<Shape>();
	private ArrayList<Shape> lowMax_list = new ArrayList<Shape>();
	private ArrayList<Shape> upMin_list = new ArrayList<Shape>();
	private ArrayList<Shape> upMax_list = new ArrayList<Shape>();
	private ArrayList<Shape> upper_Half = new ArrayList<Shape>();
	private ArrayList<Shape> lower_Half = new ArrayList<Shape>();

	private Random rand = new Random();
	private int nodes = 20;
	private Shape min0;
	private Shape max0;
	private Shape Min;
	private Shape Max;
	private boolean concave = true;

	private Color blue = new Color(0,0,255);
	private Color pink = new Color(255,20,147);
	private Color green = new Color(127,255,0);
	private Color red = new Color(255,0,0);

@Override
public void paint(Graphics g) {
	Graphics2D ga = (Graphics2D)g;

//  Initialize the nodes, give them a random position, and draw them as circles
	for (int z = 0; z < nodes; z++) {
		circle_list.add(new Ellipse2D.Float(50+rand.nextInt(800), 50+rand.nextInt(800), 10, 10));
		ga.draw(circle_list.get(z));
	}

//  Uses the maxMin method to draw the line between min0 and max0 in pink
	maxMin(ga, circle_list, pink);

//  Stores the result of maxMin as min0 and max0, and adds to the corresponding list
	min0=Min;
	max0=Max;
	upMin_list.add(min0);
	lowMax_list.add(max0);

//  Splits into upper/lower half, and removes unnecessary elements
	upper_Half = split(min0, max0, (ArrayList) circle_list.clone(), true);
	lower_Half = split(min0, max0, (ArrayList) circle_list.clone(), false);

//  Ensures that there the border is continuous in the upper half
	if (upper_Half.size()==0) {
		upper_Half.add(max0);
		upper_Half.add(min0);
	}

//  Finds max and min in the upper half, stores them in lists, split to remove
//  unnecessary elements, continue until there is no more elements in upper half
	while (upper_Half.size()>=1) {
		upper_Half = maxMin(ga, upper_Half, blue);
		upMin_list.add(Min);
		upMax_list.add(Max);
		upper_Half = split(Min, Max, upper_Half, true);
	}

//  Ensures that there the border is continuous in the lower half
	if (lower_Half.size()==0) {
		lower_Half.add(max0);
		lower_Half.add(min0);
	}
//  Finds max and min in the lower half, stores them in lists, split to remove
//  unnecessary elements, continue until there is no more elements in lower half
	while (lower_Half.size()>=1) {
		lower_Half = maxMin(ga, lower_Half, blue);
		lowMin_list.add(Min);
		lowMax_list.add(Max);
		lower_Half = split(Min, Max, lower_Half, false);
	}

//  Append the max/min elements from upper-half in the right order
    for (int z = 0; z < upMin_list.size(); z++) {
		upper_Half.add(upMin_list.get(z));
	}

	for (int z = upMax_list.size()-1; z >= 0; z--) {
		upper_Half.add(upMax_list.get(z));
	}

//  Append the max/min elements from lower-half in the right order
	for (int z = 0; z < lowMax_list.size(); z++) {
		lower_Half.add(lowMax_list.get(z));
	}

	for (int z = lowMin_list.size()-1; z >=0 ; z--) {
		lower_Half.add(lowMin_list.get(z));
	}

//  Merge and draw the lines between the border elements in red and with -2 offset
	border_list = mergeLists(lower_Half, upper_Half);
	drawBorderLines(ga, border_list, -2, red);

//  Remove concave elements from upper and lower half
	if (concave) {
		concaveRemover(min0, lower_Half, false);
		concaveRemover(max0, upper_Half, true);
	}

//  Merge and draw the lines between the border elements in green and with 0 offset
	border_list = mergeLists(lower_Half, upper_Half );
	drawBorderLines(ga, border_list, 0, green);
}

public void drawBorderLines (Graphics2D ga, ArrayList<Shape> listInput, int offset, Color color){
//  Draw lines between all the elements in a given list, connect last and first as well
//  Uses the offset to shift the lines to avoid lines on top of each other
//  Uses a given colour based on border type
	ArrayList<Shape> lineList = new ArrayList<Shape>();
	ga.setPaint(color);
	for (int z = 0; z < listInput.size(); z++) {
	if (z == (listInput.size()-1)) {

		ga.draw(new Line2D.Double(listInput.get(z).getBounds2D().getCenterX()+offset, listInput.get(z).getBounds2D().getCenterY()+offset, listInput.get(0).getBounds2D().getCenterX()+offset, listInput.get(0).getBounds2D().getCenterY()+offset));

	} else {

		lineList.add(new Line2D.Double(listInput.get(z).getBounds2D().getCenterX()+offset, listInput.get(z).getBounds2D().getCenterY()+offset, listInput.get(z+1).getBounds2D().getCenterX()+offset, listInput.get(z+1).getBounds2D().getCenterY()+offset));

		ga.draw(new Line2D.Double(listInput.get(z).getBounds2D().getCenterX()+offset, listInput.get(z).getBounds2D().getCenterY()+offset, listInput.get(z+1).getBounds2D().getCenterX()+offset, listInput.get(z+1).getBounds2D().getCenterY()+offset));

	}
	}
}

public ArrayList<Shape> split (Shape point1, Shape point2, ArrayList<Shape> list, boolean up){
//  Separates the upper and lower half based upon the boolean value "up"
//  removes all unnecessary elements, including the previous max/min
//  Manages the case where there is just one element left in the upper/lower half
	list.remove(point1);
	list.remove(point2);

	double yTemp1 = point1.getBounds2D().getCenterY();
	double yTemp2 = point2.getBounds2D().getCenterY();
	double xTemp1 = point1.getBounds2D().getCenterX();
	double xTemp2 = point2.getBounds2D().getCenterX();

	for (int z = list.size()-1; z >= 0; z--) {
		double yTemp = list.get(z).getBounds2D().getCenterY();
		double xTemp = list.get(z).getBounds2D().getCenterX();
		double yTarget =yTemp1 + (((yTemp2-yTemp1)/(xTemp2-xTemp1)) * (xTemp-xTemp1));

		if (up){
			if ( (yTemp > yTarget)) {
				list.remove(z);
			}
		}else{
			if ( (yTemp < yTarget)) {
				list.remove(z);
			}
		}
	}
	if (up){
		if (list.size() == 1) {
			upMin_list.add(list.get(0));
		}
	}else{
		if (list.size() == 1) {
			lowMax_list.add(list.get(0));
		}
	}
return list;}

public ArrayList<Shape> maxMin(Graphics2D ga, ArrayList<Shape> circleTemp_list, Color color){
//  Finds the max and min in any given list of elements
//  Draws the lines between these elements in a given colour, and with offset +2
	Min = circleTemp_list.get(0);
	Max = circleTemp_list.get(0);

	for (int z = circleTemp_list.size()-1; z >= 0 ; z--) {
		double xTemp = circleTemp_list.get(z).getBounds2D().getCenterX();
		double xTempMin = Min.getBounds2D().getCenterX();
		double xTempMax = Max.getBounds2D().getCenterX();
		if ( xTemp < xTempMin) {
			Min = circleTemp_list.get(z);
		}
		if ( xTemp > xTempMax) {
			Max = circleTemp_list.get(z);
		}
	}
	ga.setPaint(color);

	ga.draw(new Line2D.Double(Min.getBounds2D().getCenterX()+2, Min.getBounds2D().getCenterY()+2, Max.getBounds2D().getCenterX()+2, Max.getBounds2D().getCenterY()+2));

return circleTemp_list;}

public ArrayList<Shape> mergeLists (ArrayList<Shape> lowHalf, ArrayList<Shape> upHalf){
//  Combines the two lists and returns this
	ArrayList<Shape> mergedList = new ArrayList<Shape>();
	mergedList.addAll(upHalf);
	mergedList.addAll(lowHalf);
return mergedList;}

public void concaveRemover (Shape value, ArrayList<Shape> node_Half, boolean up){
//  Removes concave elements form upper or lower half based upon boolean "up"
//  Uses the relative distance from a calculated point on the line
//  between the elements neighbours.
	double xTemp1;
	double yTemp1;
	double yTarget;
	boolean ConcaveTest = true;

	while (ConcaveTest) {
		ConcaveTest = false;

		for (int z = node_Half.size()-1; z > 0; z--) {
			if (z == node_Half.size()-1) {
				xTemp1 = value.getBounds2D().getCenterX();
				yTemp1 = value.getBounds2D().getCenterY();
			} else{
				xTemp1 = node_Half.get(z+1).getBounds2D().getCenterX();
				yTemp1 = node_Half.get(z+1).getBounds2D().getCenterY();
			}
			double xTemp2 = node_Half.get(z).getBounds2D().getCenterX();
			double yTemp2 = node_Half.get(z).getBounds2D().getCenterY();
			double xTemp3 = node_Half.get(z-1).getBounds2D().getCenterX();
			double yTemp3 = node_Half.get(z-1).getBounds2D().getCenterY();

			yTarget = (yTemp1 + (((yTemp3-yTemp1)/(xTemp3-xTemp1)) * (xTemp2-xTemp1)));

			if (((yTemp2 <= yTarget) && !up) || ((yTemp2 >= yTarget) && up)) {
				node_Half.remove(z);
				ConcaveTest =true;
			}
		}
	}
}

public static void main(String args[]) {
	Frame frame = new border_finder();
	frame.setSize(900, 900);
	frame.setVisible(true);
	frame.addWindowListener(new WindowAdapter(){
		@Override
		public void windowClosing(WindowEvent we){
			System.exit(0);
		}
	});
}
}